from random import choice
from textwrap import dedent

from rich.layout import Layout
from rich.panel import Panel


class HelpState:
    keybindings, about = 0, 1


class HelpLayout(Layout):
    def __init__(self, version, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.version = version
        self.state = HelpState.keybindings

    def __call__(self, height: int):
        lines = self.text.splitlines()
        if len(lines) > height - 6:
            lines = lines[: height - 6] + ["        ..."]

        text = "\n".join(lines)
        self.update(
            Panel(
                text,
                title=(
                    "[i white]help[/i white] | [u]key bindings[/u] [dim]about"
                    if self.state == HelpState.keybindings
                    else "[i white]help[/i white] | [dim]key bindings[/dim] [u]about"
                ),
                title_align="left",
                subtitle="[dim white][u]f[/u]:fullscreen [u][][/u]:switch pane [u]?[/u]:exit help",
                subtitle_align="left",
                style="magenta",
            )
        )
        return self

    @property
    def text(self):
        """ Return the text to be displayed on the help page """
        if self.state == HelpState.keybindings:
            return dedent(
                """
                [white]
                      k ↑ - [cyan]up[/cyan]
                      j ↓ - [cyan]down[/cyan]
                        g - [cyan]go to top[/cyan]
                        G - [cyan]go to bottom[/cyan]
                l → Enter - [cyan]select[/cyan]
                    Space - [cyan]select[/cyan]
                      h ← - [cyan]go back to parent object[/cyan]
                      [ ] - [cyan]switch attribute type (public/private)[/cyan]
                      { } - [cyan]switch pane[/cyan]
                        p - [cyan]toggle full preview[/cyan]
                        d - [cyan]toggle full docstring[/cyan]
                        n - [cyan]toggle filter view[/cyan]
                        / - [cyan]open search filter[/cyan]
                      Esc - [cyan]close[/cyan]
                        c - [cyan]clear filters[/cyan]
                        o - [cyan]toggle stack view[/cyan]
                        f - [cyan]open fullscreen view[/cyan]
                        + - [cyan]increase explorer layout[/cyan]
                        - - [cyan]decrease explorer layout[/cyan]
                        = - [cyan]return explorer layout size to default[/cyan]
                        O - [cyan]open source file in [i u]$EDITOR[/i u][/cyan]
                        H - [cyan]open help page on selected attribute[/cyan]
                        i - [cyan]run [magenta]rich[/magenta][white].[/white][magenta]inspect[/magenta][white](<[/white][bright_magenta]OBJECT[/bright_magenta]>, [yellow]methods[/yellow]=[italic bright_green]True[/italic bright_green][white])[/white][/cyan]
                        I - [cyan]run [magenta]rich[/magenta][white].[/white][magenta]inspect[/magenta][white](<[/white][bright_magenta]OBJECT[/bright_magenta]>, [yellow]all[/yellow]=[italic bright_green]True[/italic bright_green][white])[/white][/cyan]
                        r - [cyan]return the selected object[/cyan]
                        ? - [cyan]toggle help page[/cyan]
                      q Q - [cyan]quit[/cyan]
                """
            ).strip()

        elif self.state == HelpState.about:
            return (
                f"""
[white]       _     _                 _
  ___ | |__ (_) _____  ___ __ | | ___  _ __ ___
 / _ \| '_ \| |/ _ \ \/ / '_ \| |/ _ \| '__/ _ \\
| (_) | |_) | |  __/>  <| |_) | | (_) | | |  __/
 \___/|_.__// |\___/_/\_\ .__/|_|\___/|_|  \___|
          |__/          |_|
Interactive Python Object Explorer

Author:          [cyan]Kyle Pollina[/cyan]
Version:         [cyan]{self.version}[/cyan]
PyPI:            [cyan]https://pypi.org/project/objexplore[/cyan]
Source:          [cyan]https://github.com/kylepollina/objexplore[/cyan][yellow italic]
Report an issue[/yellow italic]: [cyan]https://github.com/kylepollina/objexplore/issues[/cyan]\n\n"""
                + "[i]"
                + self.random_quote()
            )

    def random_quote(self):
        return choice(
            [
                "Have a nice day!!",
                "You look rather dashing today!",
                ":)",
                ":earth_africa:",
                "<3",
                "Thanks for checking this out ;)",
                "Why, hello there!",
                "Oh well look who it is!",
                "Oh if a man tried\n"
                "To take his time on Earth\n"
                "And prove before he died\n"
                "What one man's life could be worth\n"
                "I wonder what would happen to this world\n"
                " - Harry Chapin",
                "In this world, people [u i]can[/u i] change it for the better,\n"
                "and that those people who are crazy enough to think that\n"
                "they can change the world are the ones that actually do.\n"
                " - Steve Jobs",
            ]
        )


def random_error_quote():
    quotes = [
        "Aw nuts!",
        "Uh oh...",
        "Foiled again!",
        "We've been bamboozled!",
        "Something ain't right here...",
        "$#%@!!",
        "BONK!",
        "[blink]",
        "Oh no!",
        "Whoopsies.",
        "Dag nabit",
        "👀",
        "Computer over.",
        "Virus = very yes.",
        "Oh Child!",
        "I used objexplore and all I got was this lousy error message.",
    ]
    q1 = choice(quotes)
    quotes.remove(q1)
    q2 = choice(quotes)
    return q1 + " " + q2



from dataclasses import dataclass
from typing import List, Optional

import rich
from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.style import Style
from rich.text import Text
from rich.tree import Tree

from .cached_object import CachedObject
from .filter import Filter

console = Console()


@rich.repr.auto
@dataclass
class StackFrame:
    """ Datastructure to store a frame in the object stack """

    # flake8: noqa
    cached_obj: CachedObject
    filter: Filter
    state: "ExplorerState"  # type: ignore
    public_index: int
    public_window: int
    private_index: int
    private_window: int
    dict_index: int
    dict_window: int
    list_index: int
    list_window: int


class Stack:
    def __init__(self, head_obj: CachedObject):
        self.head_obj = head_obj
        self.index = 0
        self.layout = Layout(visible=False)
        self.stack: List[StackFrame] = []

    def push(self, stack_frame: StackFrame):
        self.stack.append(stack_frame)

    def pop(self) -> Optional[StackFrame]:
        if self.stack:
            return self.stack.pop()
        return None

    def __getitem__(self, item):
        return self.stack[item]

    def set_visible(self):
        self.layout.visible = True
        self.index = len(self.stack)

    def get_layout(self, width: int, current_obj: CachedObject) -> Layout:

        # Add the head obj as the base of the tree
        head_label = self.head_obj.repr
        if self.index == 0:
            head_label.style = Style(reverse=True)
        else:
            head_label.style = Style()

        head_label.truncate(width)
        stack_tree = Tree(head_label)

        # Go through the stack and add to the stack tree
        for index, stack_frame in enumerate(self.stack[1:]):
            if index + 1 == self.index:
                style = Style(reverse=True)
            else:
                style = Style()

            label = (
                Text(stack_frame.cached_obj.attr_name)
                + Text(": ")
                + stack_frame.cached_obj.typeof
            )
            label.overflow = "ellipsis"
            label.truncate(width - 4)
            stack_tree.add(
                label,
                style=style,
            )

        if self.stack != []:
            # Finally add the current obj to the tree just for clarity sake on what is being looked at
            # Only do it if the stack is not empty since we already have the head obj at the base of the tree
            if self.index == len(self.stack):
                style = Style(reverse=True)
            else:
                style = Style()
            label = Text(current_obj.attr_name) + Text(": ") + current_obj.typeof
            label.overflow = "ellipsis"
            label.truncate(width - 4)
            stack_tree.add(
                label,
                style=style,
            )

        subtitle = "[dim][u]space[/u]:select"
        if len(console.render_str(subtitle)) > width:
            subtitle = ""

        self.layout.update(
            Panel(
                stack_tree,
                title="\[stack]",
                title_align="right",
                subtitle=subtitle,
                subtitle_align="right",
                style="bright_blue",
            )
        )
        self.layout.size = len(self.stack) + 5
        return self.layout

    def move_up(self):
        if self.index > 0:
            self.index -= 1

    def move_down(self):
        if self.index < len(self.stack):
            self.index += 1

    def move_top(self):
        self.index = 0

    def move_bottom(self):
        self.index = len(self.stack) - 1

    def select(self) -> Optional[StackFrame]:
        """The stack always contains every stack frame including the current frame
        so selecting based on the index, we want to rebuild the stack to be everything
        up to the selected index and then pop the very top off and return

        TODO more documentation
        """
        if self.index != len(self.stack):
            self.stack = self.stack[: self.index + 1]
            if self.stack:
                return self.stack.pop()
        return None



from typing import List

import blessed
import rich
from blessed import Terminal
from blessed.keyboard import Keystroke
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.layout import Layout
from rich.panel import Panel
from rich.style import Style
from rich.text import Text

from .cached_object import CachedObject
from .config import box_type

console = Console()
highlighter = ReprHighlighter()

# TODO scroll search if input longer than panel width


def isclass(cached_obj: CachedObject):
    return cached_obj.isclass


def isfunction(cached_obj: CachedObject):
    return cached_obj.isfunction


def ismethod(cached_obj: CachedObject):
    return cached_obj.ismethod


def ismodule(cached_obj: CachedObject):
    return cached_obj.ismodule


def isbuiltin(cached_obj: CachedObject):
    return cached_obj.isbuiltin


def isint(cached_obj: CachedObject):
    return type(cached_obj.obj) == int


def isstr(cached_obj: CachedObject):
    return type(cached_obj.obj) == str


def isfloat(cached_obj: CachedObject):
    return type(cached_obj.obj) == float


def isbool(cached_obj: CachedObject):
    return type(cached_obj.obj) == bool


def isdict(cached_obj: CachedObject):
    return type(cached_obj.obj) == dict


def islist(cached_obj: CachedObject):
    return type(cached_obj.obj) == list


def istuple(cached_obj: CachedObject):
    return type(cached_obj.obj) == tuple


def isset(cached_obj: CachedObject):
    return type(cached_obj.obj) == set


@rich.repr.auto
class Filter:
    def __init__(self, term: Terminal):
        self.term = term
        self.layout = Layout(visible=False)

        self.filters = {
            "class": [False, isclass],
            "function": [False, isfunction],
            "method": [False, ismethod],
            "module": [False, ismodule],
            "int": [False, isint],
            "str": [False, isstr],
            "float": [False, isfloat],
            "bool": [False, isbool],
            "dict": [False, isdict],
            "list": [False, islist],
            "tuple": [False, istuple],
            "set": [False, isset],
            "builtin": [False, isbuiltin],
        }
        self.index = 0
        self.receiving_input = False
        self.search_filter = ""
        self.cursor_pos = 0
        self.key_history: List[Keystroke] = []

    def move_down(self):
        if self.index < len(self.filters) - 1:
            self.index += 1

    def move_up(self):
        if self.index > 0:
            self.index -= 1

    def move_top(self):
        self.index = 0

    def move_bottom(self):
        self.index = len(self.filters) - 1

    def get_enabled_filters(self) -> list:
        return [
            function
            for name, (enabled, function) in self.filters.items()
            if enabled is True
        ]

    @property
    def selected_filter(self):
        return list(self.filters.keys())[self.index]

    def toggle(self, cached_obj: CachedObject):
        """ Toggle the selected filter on or off and update the cached_obj filters with the new filters """
        self.filters[self.selected_filter][0] = not self.filters[self.selected_filter][
            0
        ]
        cached_obj.set_filters(self.get_enabled_filters(), self.search_filter)

    def clear_filters(self, cached_obj: CachedObject):
        for name, filter_data in self.filters.copy().items():
            self.filters[name][0] = False
        self.search_filter = ""
        self.cursor_pos = 0
        cached_obj.set_filters([])

    def get_lines(self) -> List[Text]:
        lines = []
        for index, (name, (enabled, method)) in enumerate(self.filters.items()):
            line = (
                Text("[", style=Style(color="white"))
                + Text("X" if enabled else " ", style=Style(color="blue"))
                + Text("] ", style=Style(color="white"))
                + Text(name, style=Style(color="magenta"))
            )
            if index == self.index:
                line.style += Style(reverse=True)  # type: ignore
            lines.append(line)

        if self.search_filter:
            lines.extend(
                [
                    Text("Search filter:", style=Style(italic=True, underline=True)),
                    Text(" " + self.search_filter),
                ]
            )

        return lines

    def add_search_char(
        self,
        key: blessed.keyboard.Keystroke,
        cached_obj: CachedObject,
        live_update: bool,
    ):
        self.key_history.append(key)
        self.search_filter = (
            self.search_filter[: self.cursor_pos]
            + str(key)
            + self.search_filter[self.cursor_pos :]
        )
        self.cursor_pos += 1

        if live_update:
            cached_obj.set_filters(self.get_enabled_filters(), self.search_filter)

    def backspace(self, cached_obj: CachedObject, live_update: bool):
        """Delete the character before the cursor.
        Args:
            cached_obj: The current object being explored.
            live_update: True/False value whether to update the cached_obj
                filters. If there are too many attributes then updating the
                filters will slow down.
        """
        if self.cursor_pos == 0 and self.search_filter == "":
            self.cancel_search(cached_obj)
        # if the cursor is at the beginning but there is still text in the search, do nothing
        elif self.cursor_pos == 0 and self.search_filter:
            return

        self.search_filter = (
            self.search_filter[: self.cursor_pos - 1]
            + self.search_filter[self.cursor_pos :]
        )
        self.cursor_left()

        if live_update:
            cached_obj.set_filters(self.get_enabled_filters(), self.search_filter)

    def cancel_search(self, cached_obj: CachedObject):
        self.search_filter = ""
        self.cursor_pos = 0
        self.layout.visible = False
        self.receiving_input = False
        cached_obj.set_filters(self.get_enabled_filters(), self.search_filter)

    def cursor_left(self):
        if self.cursor_pos > 0:
            self.cursor_pos -= 1

    def cursor_right(self):
        if self.cursor_pos < len(self.search_filter):
            self.cursor_pos += 1

    def end_search(self, cached_obj: CachedObject):
        self.receiving_input = False
        self.layout.visible = False
        cached_obj.set_filters(
            self.get_enabled_filters(), search_filter=self.search_filter
        )

    def get_layout(self, width: int) -> Layout:
        if self.receiving_input:
            return self.get_input_layout()

        subtitle = "[dim][u]c[/u]:clear [u]space[/u]:select"
        if width <= 25:
            subtitle = "[dim][u]space[/u]:select"
        if len(console.render_str(subtitle)) > width:
            subtitle = ""

        lines = self.get_lines()
        self.layout.update(
            Panel(
                Text("\n").join(lines),
                title="\[filter]",
                title_align="right",
                subtitle=subtitle,
                subtitle_align="right",
                style="bright_magenta",
                box=box_type,
            )
        )
        self.layout.size = len(lines) + 2
        return self.layout

    def get_input_layout(self) -> Layout:
        if len(self.search_filter) == 0:
            search_text = Text(
                "█", style=Style(underline=True, blink=True, reverse=True)
            )
        elif self.cursor_pos == len(self.search_filter):
            search_text = Text(self.search_filter) + Text(
                "█", style=Style(underline=True, blink=True, reverse=True)
            )
        else:
            search_text = (
                Text(self.search_filter[: self.cursor_pos])
                + Text(
                    self.search_filter[self.cursor_pos],
                    style=Style(
                        underline=True,
                        blink=True,
                        color="black",
                        bgcolor="aquamarine1",
                    ),
                )
                + Text(self.search_filter[self.cursor_pos + 1 :])
            )

        self.layout.update(
            Panel(
                search_text,
                title="\[search]",
                title_align="right",
                subtitle="[dim][u]esc[/u]:cancel",
                subtitle_align="right",
                style=Style(color="aquamarine1"),
                box=box_type,
            )
        )
        self.layout.size = 3
        return self.layout



def is_empty(obj):
    """ Check to see if the object is equal to any of the following objects """
    return not any(obj is x for x in [None, [], (), {}, set()])
    # try:
    #     return not any(obj is x for x in [None, [], (), {}, set()])
    # except Exception:
    #     # TODO this might not be needed anymore
    #     return True



from typing import Union

from blessed import Terminal
from rich.layout import Layout
from rich.panel import Panel
from rich.pretty import Pretty
from rich.style import Style
from rich.syntax import Syntax
from rich.text import Text

from .cached_object import CachedObject
from .help_layout import HelpLayout
from .config import box_type


class OverviewState:
    all, docstring, value = range(3)


class PreviewState:
    repr, source = range(2)


class Overview:
    def __init__(self, term: Terminal, version: str):
        self.term = term
        self.layout = Layout()
        self.help_layout = HelpLayout(version, visible=False, ratio=3)
        self.state = OverviewState.all
        self.preview_state = PreviewState.repr

    @property
    def layout_width(self):
        return (self.term.width - 2) // 4 * 3

    def get_layout(self, cached_obj: CachedObject) -> Layout:
        """
        :param cached_obj: The selected cached object given by the explorer layout
        """
        if self.help_layout.visible:
            return self.help_layout(self.term.height)

        elif self.state == OverviewState.docstring:
            self.layout.update(
                self.get_docstring_panel(
                    cached_obj=cached_obj,
                    term_height=self.term.height,
                )
            )
            return self.layout

        elif self.state == OverviewState.value:
            self.layout.update(self.get_value_panel(cached_obj))
            return self.layout

        elif self.state == OverviewState.all:
            layout = Layout()
            layout.split_column(
                Layout(self.get_value_panel(cached_obj)),
                self.get_info_layout(cached_obj),
                Layout(
                    self.get_docstring_panel(
                        cached_obj=cached_obj, term_height=self.term.height
                    ),
                ),
            )
            return layout
        else:
            raise ValueError("Unexpected overview state")

    def get_value_panel(self, cached_obj: CachedObject):
        renderable: Union[str, Pretty, Syntax]
        if not callable(cached_obj.obj):
            title = "[i]preview[/i] | [i][cyan]repr[/cyan]()[/i]"
            subtitle = "[dim][u]p[/u]:toggle [u]f[/u]:fullscreen [u]{}[/u]:switch pane"
            renderable = cached_obj.pretty

            if self.state == OverviewState.all:
                renderable.max_length = max((self.term.height - 6) // 2 - 7, 1)
            else:
                renderable.max_length = max(self.term.height - 9, 1)

        else:
            if self.preview_state == PreviewState.repr:
                renderable = cached_obj.pretty
                title = "[i]preview[/i] | [i][cyan]repr[/cyan]()[/i] [dim]source"

            if self.preview_state == PreviewState.source:
                renderable = cached_obj.get_source(self.term.height)
                title = (
                    "[i]preview[/i] | [dim][cyan]repr[/cyan]()[/dim] [underline]source"
                )

            subtitle = "[dim][u]p[/u]:toggle [u]f[/u]:fullscreen [u]{}[/u]:switch pane"

        return Panel(
            renderable,
            title=title,
            title_align="left",
            subtitle=subtitle,
            subtitle_align="left",
            style=Style(color="white"),
            box=box_type,
        )

    def get_info_layout(self, cached_obj: CachedObject):
        if cached_obj.length is not None:
            layout = Layout(size=3)
            layout.split_row(
                Layout(self.get_type_panel(cached_obj)),
                Layout(
                    Panel(
                        str(cached_obj.length),
                        title="[i][cyan]len[/cyan]()[/i]",
                        title_align="left",
                        style="white",
                        box=box_type,
                    )
                ),
            )
            return layout

        else:
            return self.get_type_panel(cached_obj)

    def get_type_panel(self, cached_obj: CachedObject):
        return Layout(
            Panel(
                cached_obj.typeof,
                title="[i][cyan]type[/cyan]()[/i]",
                title_align="left",
                style="white",
                box=box_type,
            ),
            size=3,
        )

    def get_docstring_panel(
        self,
        cached_obj: CachedObject,
        term_height: int,
    ) -> Panel:
        """ Build the docstring panel """
        title = "[i]docstring"
        if self.state == OverviewState.docstring:
            subtitle = "[dim][u]d[/u]:toggle [u]f[/u]:fullscreen"
        else:
            subtitle = "[dim][u]d[/u]:toggle"
        docstring = Text("\n").join(cached_obj.docstring_lines[:term_height])
        return Panel(
            docstring,
            title=title,
            title_align="left",
            subtitle=subtitle,
            subtitle_align="left",
            style="white",
            box=box_type,
        )



import inspect
import os
import pydoc
import signal
import subprocess
import time
from typing import Any, Optional, Union

import rich
from blessed import Terminal
from blessed.keyboard import Keystroke
from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.style import Style
from rich.syntax import Syntax
from rich.text import Text

from .cached_object import CachedObject
from .explorer import Explorer, ExplorerState
from .help_layout import HelpState, random_error_quote
from .overview import Overview, OverviewState, PreviewState
from .config import box_type

# TODO object highlighted on stack view should be shown on the overview
# TODO support ctrl-a + (whatever emacs keybinding to go to end of line)
#  https://www.gnu.org/software/bash/manual/html_node/Commands-For-Moving.html
# TODO builtin frame/stack explorer? from objexplore import stackexplore


version = "1.6.2"
console = Console()
EDITOR = os.environ.get("EDITOR")


class ObjExploreApp:
    """ Main Application class """

    error_style = Style(color="red")
    main_style = Style(color="blue")
    term = Terminal()

    def __init__(self, obj: Any, name: str):
        cached_obj = CachedObject(obj, attr_name=name)
        # Figure out all the attributes of the current obj's attributes
        cached_obj.cache()

        self.explorer = Explorer(term=self.term, cached_obj=cached_obj)
        self.overview = Overview(term=self.term, version=version)

        # Run self.draw() whenever the win change signal is caught
        try:
            signal.signal(signal.SIGWINCH, self.draw)
        # Windows does not have SIGWINCH signal
        except AttributeError:
            pass

    def explore(self) -> Optional[Any]:
        """ Open the interactive explorer. This is the main running loop """

        key = None
        res = None

        # Clear the screen
        print(self.term.clear, end="")

        with self.term.cbreak(), self.term.hidden_cursor():
            while True:
                try:
                    self.draw()
                    key = self.term.inkey()
                    self.process_key_event(key)

                except RuntimeError as err:
                    # Some kind of error thrown during resizing events. Ignore and continue
                    if (
                        err.args[0]
                        == "reentrant call inside <_io.BufferedWriter name='<stdout>'>"
                    ):
                        pass
                    # Otherwise it is a new error. Raise
                    else:
                        raise err
                except StopIteration:
                    if key == "r":
                        res = self.explorer.selected_object.obj
                    break

        # Unhide the cursor
        print("\x1b[?25h", end="")

        return res

    def process_key_event(self, key: Keystroke) -> Any:
        """ Process the incoming key """

        if self.explorer.filter.receiving_input:
            if key.code == self.term.KEY_BACKSPACE:
                self.explorer.filter.backspace(
                    cached_obj=self.explorer.cached_obj,
                    live_update=self.explorer.live_update,
                )
            elif key.code == self.term.KEY_ESCAPE:
                self.explorer.filter.cancel_search(self.explorer.cached_obj)
            elif key.code == self.term.KEY_ENTER:
                self.explorer.filter.end_search(cached_obj=self.explorer.cached_obj)
            elif key.code == self.term.KEY_LEFT:
                self.explorer.filter.cursor_left()
            elif key.code == self.term.KEY_RIGHT:
                self.explorer.filter.cursor_right()
            elif key.code in (self.term.KEY_UP, self.term.KEY_DOWN):
                return
            else:
                self.explorer.filter.add_search_char(
                    key=key,
                    cached_obj=self.explorer.cached_obj,
                    live_update=self.explorer.live_update,
                )
            return

        if key in ("q", "Q", "r"):
            raise StopIteration

        # Help page ###########################################################

        if self.overview.help_layout.visible:
            # Close help page
            if key == "?" or key.code in (
                self.term.KEY_ESCAPE,
                self.term.KEY_BACKSPACE,
            ):
                self.overview.help_layout.visible = False
                return

            # Fullscreen
            elif key == "f":
                with console.capture() as capture:
                    console.print(self.overview.help_layout.text)
                str_out = capture.get()
                pydoc.pager(str_out)
                return

            # Switch panes
            elif key in ("{", "}", "[", "]"):
                if self.overview.help_layout.state == HelpState.keybindings:
                    self.overview.help_layout.state = HelpState.about
                elif self.overview.help_layout.state == HelpState.about:
                    self.overview.help_layout.state = HelpState.keybindings
                return

            elif key in ("j", "k", "o", "n") or key.code in (
                self.term.KEY_UP,
                self.term.KEY_DOWN,
            ):
                # Continue on and process these keys as normal
                self.overview.help_layout.visible = False
                # Do not return so we can run the key action after the help window has closed
            else:
                return

        if key == "?" and self.overview.help_layout.visible is False:
            self.overview.help_layout.visible = True
            return

        # Stack ###############################################################

        if key == "o":
            if self.explorer.stack.layout.visible:
                self.explorer.stack.layout.visible = False
            elif self.explorer.filter.layout.visible:
                self.explorer.filter.layout.visible = False
                self.explorer.stack.set_visible()
            else:
                self.explorer.stack.set_visible()

        elif (
            key.code in (self.term.KEY_BACKSPACE, self.term.KEY_ESCAPE)
            and self.explorer.stack.layout.visible
        ):
            self.explorer.stack.layout.visible = False

        elif (
            key == " " or key.code == self.term.KEY_ENTER
        ) and self.explorer.stack.layout.visible:
            self.explorer.explore_selected_stack_object()

        elif (
            key == "j" or key.code == self.term.KEY_DOWN
        ) and self.explorer.stack.layout.visible:
            self.explorer.stack.move_down()

        elif (
            key == "k" or key.code == self.term.KEY_UP
        ) and self.explorer.stack.layout.visible:
            self.explorer.stack.move_up()

        elif key == "g" and self.explorer.stack.layout.visible:
            self.explorer.stack.move_top()

        elif key == "G" and self.explorer.stack.layout.visible:
            self.explorer.stack.move_bottom()

        # disable these keys when the stack explorer is visible
        elif (
            key in ("l", "[", "]", "{", "}", "h") and self.explorer.stack.layout.visible
        ):
            return

        # Filter ##############################################################

        elif key == "n":
            if self.explorer.filter.layout.visible:
                self.explorer.filter.layout.visible = False
            elif self.explorer.stack.layout.visible:
                self.explorer.stack.layout.visible = False
                self.explorer.filter.layout.visible = True
            else:
                self.explorer.filter.layout.visible = True

        elif key == "/" and not self.explorer.filter.receiving_input:
            self.explorer.stack.layout.visible = False
            self.explorer.filter.receiving_input = True
            self.explorer.filter.layout.visible = True

        elif (
            key == " " or key.code == self.term.KEY_ENTER
        ) and self.explorer.filter.layout.visible:
            self.explorer.filter.toggle(self.explorer.cached_obj)
            self.explorer.reset_index()

        elif (
            key.code in (self.term.KEY_ESCAPE, self.term.KEY_BACKSPACE)
        ) and self.explorer.filter.layout.visible:
            self.explorer.filter.layout.visible = False

        elif (
            key == "j" or key.code == self.term.KEY_DOWN
        ) and self.explorer.filter.layout.visible:
            self.explorer.filter.move_down()

        elif (
            key == "k" or key.code == self.term.KEY_UP
        ) and self.explorer.filter.layout.visible:
            self.explorer.filter.move_up()

        elif key == "g" and self.explorer.filter.layout.visible:
            self.explorer.filter.move_top()

        elif key == "G" and self.explorer.filter.layout.visible:
            self.explorer.filter.move_bottom()

        elif key == "c":
            self.explorer.filter.clear_filters(self.explorer.cached_obj)

        # Explorer ############################################################

        elif key == "k" or key.code == self.term.KEY_UP:
            self.explorer.move_up()

        elif key == "j" or key.code == self.term.KEY_DOWN:
            self.explorer.move_down()

        elif key in ("l") or key.code in (
            self.term.KEY_ENTER,
            self.term.KEY_RIGHT,
            self.term.KEY,
        ):
            self.explorer.explore_selected_object()

        # Go back to parent
        elif (
            key == "h" or key.code == self.term.KEY_LEFT
        ) and self.explorer.stack.stack:
            self.explorer.explore_parent_obj()

        elif key == "g":
            self.explorer.move_top()

        elif key == "G":
            self.explorer.move_bottom()

        # Switch between public and private attributes
        elif key in ("[", "]"):
            if self.explorer.state == ExplorerState.public:
                self.explorer.state = ExplorerState.private

            elif self.explorer.state == ExplorerState.private:
                self.explorer.state = ExplorerState.public

        elif key == "+":
            self.explorer.increase_width()

        elif key in ("_", "-"):
            self.explorer.decrease_width()

        elif key == "=":
            self.explorer.extra_width = 0

        # Overview ############################################################

        elif key in ("{", "}"):
            if not callable(self.explorer.selected_object.obj):
                return

            if self.overview.preview_state == PreviewState.repr:
                self.overview.preview_state = PreviewState.source
            elif self.overview.preview_state == PreviewState.source:
                self.overview.preview_state = PreviewState.repr

        # Toggle docstring view
        elif key == "d":
            self.overview.state = (
                OverviewState.docstring
                if self.overview.state != OverviewState.docstring
                else OverviewState.all
            )

        # Toggle value view
        elif key == "p":
            self.overview.state = (
                OverviewState.value
                if self.overview.state != OverviewState.value
                else OverviewState.all
            )

        # Fullscreen
        elif key == "f":
            printable: Union[str, Syntax, Text]

            if self.overview.state == OverviewState.docstring:
                printable = self.explorer.selected_object.docstring

            elif self.overview.preview_state == PreviewState.repr:
                printable = self.explorer.selected_object.obj

            elif (
                self.overview.preview_state == PreviewState.source
                and self.explorer.selected_object._source
            ):
                printable = self.explorer.selected_object.get_source(fullscreen=True)

            else:
                printable = self.explorer.selected_object.obj

            with console.capture() as capture:
                console.print(printable)
            str_out = capture.get()
            pydoc.pager(str_out)

        elif key == "O":
            try:
                path = inspect.getabsfile(self.explorer.selected_object.obj)
                subprocess.call([EDITOR, path])  # type: ignore
                # Re-hide the cursor
                print("\x1b[?25l", end="")
            except Exception:
                self.error()

        elif key == "H":
            help(self.explorer.selected_object.obj)

        elif key == "i":
            with console.capture() as capture:
                rich.inspect(
                    self.explorer.selected_object.obj,
                    console=console,
                    methods=True,
                )
            str_out = capture.get()
            pydoc.pager(str_out)

        elif key == "I":
            with console.capture() as capture:
                rich.inspect(
                    self.explorer.selected_object.obj, console=console, all=True
                )
            str_out = capture.get()
            pydoc.pager(str_out)

    def draw(self, *_):
        """ Draw the application. the *_ argument is due to resize events and are unused """
        print(self.term.home, end="")
        layout = Layout()
        layout.split_row(
            self.explorer.get_layout(),
            self.overview.get_layout(self.explorer.selected_object),
        )

        title = (
            self.explorer.cached_obj.dotpath
            + Text(" | ", style="white")
            + self.explorer.cached_obj.typeof
        )

        object_explorer = Panel(
            layout,
            title=title,
            subtitle=(
                "[red][u]q[/u]:quit[/red] "
                "[cyan][u]?[/u]:help[/] "
                "[bright_blue][u]o[/u]:stack [/bright_blue][bright_magenta][u]n[/u]:filter [/bright_magenta][aquamarine1][u]/[/u]:search [/aquamarine1][u]r[/u]:return"
            ),
            subtitle_align="left",
            height=self.term.height - 1,
            style=self.main_style,
            box=box_type,
        )
        rich.print(object_explorer, end="")

    def error(self):
        """ Color the outside red and pause for a split second """
        self.main_style = self.error_style
        self.draw()
        time.sleep(0.25)
        self.main_style = self.main_style


def explore(obj: Any) -> Any:
    """
    Run the explorer on the given object

    Get the name of the variable sent to this function
    If someone calls this function like:
    >>> df = pandas.DataFrame()
    >>> explore(df)
    Then we want to extract the string 'df' to the variable `name`
    Depending on if explore is called from a script or from the REPL,
    `name` may or may not refer to the correct object
    If called from the REPL:
    >>> x = 'hello'
    >>> y = 'world'
    >>> explore(x)
    then name == 'x'
    If called from a script:
    python -c "x = 'hello'; y = 'world'; import objexplore; objexplore.explore(x)"
    then name == 'y'

    I dont know of any way to fix this
    """

    frame = inspect.currentframe()
    name = frame.f_back.f_code.co_names[1]  # type: ignore
    app = ObjExploreApp(obj, name=name)
    try:
        return app.explore()

    except Exception as err:
        print(app.term.move_down(app.term.height))
        console.print_exception(show_locals=True)
        print()
        rich.print(f"[red]{random_error_quote()}")
        formatted_link = f"https://github.com/kylepollina/objexplore/issues/new?assignees=&labels=&template=bug_report.md&title={err}".replace(
            " ", "+"
        )
        print("Please report the issue here:")
        rich.print(f"   [link={formatted_link}][u]{formatted_link}[/u][/link]")
        print()
        rich.print(
            "[yellow italic]Make sure to copy/paste the above traceback to the issue page to make this quicker to fix :)"
        )



from pathlib import Path
from setuptools import setup

from objexplore import version as VERSION

# The directory containing this file
cur_dir = Path(__file__).parent

# The text of the README file
README = (cur_dir / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="objexplore",
    version=VERSION,
    description="Interactive Python Object Explorer",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/kylepollina/objexplore",
    author="Kyle Pollina",
    author_email="kylepollina@pm.me",
    license="",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=["objexplore"],
    include_package_data=True,
    install_requires=["blessed==1.17.12", "rich>=10.9.0"],
)



from dataclasses import dataclass
import importlib
import inspect
import pkgutil
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.pretty import Pretty
from rich.style import Style
from rich.syntax import Syntax
from rich.text import Text

from .utils import is_empty

highlighter = ReprHighlighter()


PUBLIC = "PUBLIC"
PRIVATE = "PRIVATE"

console = Console()


def safegetattr(obj, attr):
    try:
        return getattr(obj, attr)
    except Exception:
        return None


class CachedObject:
    """Internal representation of every object that is being inspected/explored by objexplore

    TODO add documentation on all the attributes of this object
    TODO look up how other libraries document thier attributes
    """

    def __init__(
        self,
        obj: Any,
        parent_path: Text = None,
        attr_name: str = None,
        index: Any = None,
        hidden: bool = False,
    ):
        self.obj = obj
        self.is_callable = callable(obj)
        self.attr_name = attr_name if attr_name else repr(self.obj)

        if self.obj is None:
            # TODO this doesn't seem like the right choice but removing it causes a crash. Investigate!
            self.dotpath = highlighter("None")

        elif attr_name is not None:
            if not parent_path:
                self.dotpath = Text(attr_name, style=Style(color="cyan"))
            else:
                self.dotpath = (
                    parent_path
                    + Text(".", style=Style(color="white"))
                    + Text(attr_name, style=Style(color="cyan"))
                )

        elif index is not None:
            if type(index) == str:
                repr_index = console.render_str(f'"{index}"')
            else:
                repr_index = console.render_str(str(index))
            if not parent_path:
                self.dotpath = (
                    Text("[", style=Style(color="white"))
                    + repr_index
                    + Text("]", style=Style(color="white"))
                )
            else:
                self.dotpath = (
                    parent_path
                    + Text("[", style=Style(color="white"))
                    + repr_index
                    + Text("]", style=Style(color="white"))
                )
        else:
            raise ValueError("Need to specify an attribute name or an index")

        self.plain_attrs = dir(self.obj)

        if "__weakref__" in self.plain_attrs:
            # Ignore weakrefs
            # Why??? I don't remember
            self.plain_attrs.remove("__weakref__")

        self.plain_public_attributes = sorted(
            attr for attr in self.plain_attrs if not attr.startswith("_")
        )
        self.plain_private_attributes = sorted(
            attr for attr in self.plain_attrs if attr.startswith("_")
        )

        self.public_attributes: Dict[str, CachedObject] = {}
        self.private_attributes: Dict[str, CachedObject] = {}
        self.filtered_public_attributes: Dict[str, CachedObject] = {}
        self.filtered_private_attributes: Dict[str, CachedObject] = {}

        try:
            self._source = inspect.getsource(self.obj)  # type: ignore
        except Exception:
            self._source = ""

        self.length: Optional[int]

        try:
            self.length = len(self.obj)  # type: ignore
        except TypeError:
            self.length = None

        self.isbuiltin: bool = inspect.isbuiltin(self.obj)
        self.isclass: bool = inspect.isclass(self.obj)
        self.isfunction: bool = inspect.isfunction(self.obj)
        self.ismethod: bool = inspect.ismethod(self.obj)
        self.ismethoddescriptor: bool = inspect.ismethoddescriptor(self.obj)
        self.ismodule: bool = inspect.ismodule(self.obj)

        self.filters: List[Union[bool, Callable[[Any], Any]]] = []
        self.search_filter: str = ""

        # Highlighted attributes
        self.typeof: Text = highlighter(str(type(self.obj)))
        self.docstring: Text = console.render_str(inspect.getdoc(self.obj) or "None")
        self.docstring_lines = self.docstring.split()
        self.repr = highlighter(repr(self.obj))
        if "\n" in self.repr:
            self.repr = self.repr.split("\n")[0]
        self.repr.overflow = "ellipsis"
        self.pretty = Pretty(self.obj)

        self.text = Text(self.attr_name, style=Style(), overflow="ellipsis")

        if self.ismodule:
            self.text.style = Style(color="blue")
        elif self.isclass:
            self.text.style = Style(color="magenta")
        elif (
            self.isfunction
            or self.ismethod
            or self.ismethoddescriptor
            # builtin_function_or_method type. Don't know where this is defined
            or isinstance(self.obj, type("".capitalize))
        ):
            self.text.style = Style(color="cyan", italic=True)
            self.text += Text("()", style=Style(color="white"))
        elif type(self.obj) == dict:
            self.text.style = Style(color="light_sea_green")
            self.text = (
                Text("{**", style=Style(color="white"))
                + self.text
                + Text("}", style=Style(color="white"))
            )
        elif type(self.obj) == list:
            self.text.style = Style(color="indian_red1")
            self.text = (
                Text("[*", style=Style(color="white"))
                + self.text
                + Text("]", style=Style(color="white"))
            )
        elif type(self.obj) == tuple:
            self.text.style = Style(color="pale_violet_red1")
            self.text = (
                Text("(*", style=Style(color="white"))
                + self.text
                + Text(")", style=Style(color="white"))
            )
        elif type(self.obj) == set:
            self.text.style = Style(color="light_goldenrod3")
            self.text = (
                Text("{*", style=Style(color="white"))
                + self.text
                + Text("}", style=Style(color="white"))
            )

        if not is_empty(self.obj):
            self.text.style += Style(dim=True, strike=True)  # type: ignore

        if hidden:
            self.text.style += Style(dim=True)  # type: ignore

    @property
    def title(self):
        """ TODO """
        # for cases when the object is a huge dictionary we shouldnt try to render the whole dict
        if len(self.repr.plain) > console.width - 4:
            return Text(self.attr_name) + Text(" ") + self.typeof
        title = self.repr.copy()
        title.truncate(console.width - 4)
        return title

    def cache(self):
        """ Cache any attributes that are useful to this object for easy access later """

        if not self.public_attributes:
            for attr in self.plain_public_attributes:
                self.public_attributes[attr] = CachedObject(
                    safegetattr(self.obj, attr),
                    parent_path=self.dotpath,
                    attr_name=attr,
                )

        if not self.private_attributes:
            for attr in self.plain_private_attributes:
                self.private_attributes[attr] = CachedObject(
                    safegetattr(self.obj, attr),
                    parent_path=self.dotpath,
                    attr_name=attr,
                )

        # Sometimes a module will have submodules that are not referenced from a call to `dir()`
        # This check will look through all submodules that are not referenced by `dir()` and add
        # them to the cached attributes
        if self.ismodule:
            prefix = safegetattr(self.obj, "__name__") + "."
            path = safegetattr(self.obj, "__path__")
            for importer, full_module_name, ispkg in pkgutil.iter_modules(path, prefix):
                name = full_module_name.rsplit(".")[-1]
                if name in self.public_attributes or name in self.private_attributes:
                    # Skip over submodules that have already been indexed
                    continue

                try:
                    # If we have not encountered this module, try to import it
                    module = importlib.import_module(full_module_name)
                except Exception:
                    continue

                if not name.startswith("_"):
                    self.public_attributes[name] = CachedObject(
                        module, parent_path=self.dotpath, attr_name=name, hidden=True
                    )
                else:
                    self.private_attributes[name] = CachedObject(
                        module, parent_path=self.dotpath, attr_name=name, hidden=True
                    )

        self.num_public_attributes: int = len(self.public_attributes)
        self.num_private_attributes: int = len(self.private_attributes)

        self.filter()

    def set_filters(
        self, filters: List[Union[bool, Callable[[Any], Any]]], search_filter: str = ""
    ):
        """ Reset the filters associated with this object, and rerun the filtering process again with the new filters """
        self.filters = filters
        self.search_filter = search_filter.lower()
        self.filter()

    def filter(self):
        """ Run the filters on all of this objects attributes """
        self.filtered_public_attributes = {}
        for attr, cached_obj in self.public_attributes.items():
            if self.search_filter not in attr.lower():
                continue
            if not self.filters:
                self.filtered_public_attributes[attr] = cached_obj
            else:
                # Only keep objects that match the filter
                for _filter in self.filters:
                    if _filter(cached_obj):
                        self.filtered_public_attributes[attr] = cached_obj
                        break
        self.num_filtered_public_attributes = len(self.filtered_public_attributes)

        self.filtered_private_attributes = {}
        for attr, cached_obj in self.private_attributes.items():
            if self.search_filter not in attr.lower():
                continue
            if not self.filters:
                self.filtered_private_attributes[attr] = cached_obj
            else:
                # Only keep objects that match the filter
                for _filter in self.filters:
                    if _filter(cached_obj):
                        self.filtered_private_attributes[attr] = cached_obj
                        break
        self.num_filtered_private_attributes = len(self.filtered_private_attributes)

        self.filtered_dict: Dict[str, FilteredDictKey] = {}
        if type(self.obj) == dict:
            for key, val in self.obj.items():
                repr_key: Text
                repr_val: Text

                if type(key) == str:
                    repr_key = console.render_str(f'"{key}"')
                elif type(key) in (int, float, dict, list, set, tuple, bool, None):
                    repr_key = console.render_str(str(key))
                else:
                    repr_key = highlighter(str(key))

                repr_val = highlighter(str(type(val)))

                if not is_empty(val):
                    repr_val.style += " dim"
                    repr_val.style = repr_val.style.strip()

                line = Text(" ") + repr_key + Text(": ") + repr_val
                line.overflow = "ellipsis"

                cached_obj = CachedObject(val, parent_path=self.dotpath, index=key)

                if type(key) == str and self.search_filter not in key.lower():
                    continue
                if self.filters:
                    for _filter in self.filters:
                        if _filter(cached_obj):
                            self.filtered_dict[key] = FilteredDictKey(
                                text=line, cached_object=cached_obj
                            )
                            break
                else:
                    self.filtered_dict[key] = FilteredDictKey(
                        text=line, cached_object=cached_obj
                    )

        self.num_filtered_dict_keys = len(self.filtered_dict)

        self.filtered_list: List[Tuple[Text, CachedObject]] = []
        if isinstance(self.obj, (list, tuple, set)):
            for index, item in enumerate(self.obj):
                line = (
                    Text(" [", style=Style(color="white"))
                    + Text(str(index), style=Style(color="blue"))
                    + Text("] ", style=Style(color="white"))
                    + highlighter(str(type(item)))
                )
                if not is_empty(item):
                    line.style += Style(dim=True)

                self.filtered_list.append(
                    (line, CachedObject(item, parent_path=self.dotpath, index=index))
                )
            if self.filters:
                new_filtered_list: List[Tuple[Text, CachedObject]] = []
                for line, cached_obj in self.filtered_list:
                    for _filter in self.filters:
                        if _filter(cached_obj):
                            new_filtered_list.append((line, cached_obj))
                            break
                self.filtered_list = new_filtered_list
        self.num_filtered_list_items = len(self.filtered_list)

    def current_visible_attributes(self):
        """ TODO """
        if self.filtered_dict:
            return self.filtered_dict
        elif self.filtered_list:
            return self.filtered_list

    def get_source(
        self, term_height: int = 0, fullscreen: bool = False
    ) -> Union[Syntax, str]:
        """ TODO """
        if not fullscreen and not term_height:
            raise ValueError("Need a terminal height")

        if not self._source:
            return "[red italic]Source code unavailable"

        if fullscreen:
            return Syntax(
                self._source, "python", line_numbers=True, background_color="default"
            )
        else:
            return Syntax(
                self._source,
                "python",
                line_numbers=True,
                line_range=(0, term_height),
                background_color="default",
            )


@dataclass
class FilteredDictKey:
    """ TODO """

    text: Text
    cached_object: CachedObject

    def __iter__(self):
        # return [self.text, self.cached_object]
        yield self.text
        yield self.cached_object



from typing import Optional

from blessed import Terminal
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.layout import Layout
from rich.panel import Panel
from rich.style import Style
from rich.text import Text

from .cached_object import CachedObject
from .filter import Filter
from .stack import Stack, StackFrame
from .config import box_type

console = Console()

highlighter = ReprHighlighter()


class ExplorerState:
    public = "ExplorerState.public"
    private = "ExplorerState.private"
    dict = "ExplorerState.dict"
    list = "ExplorerState.list"
    tuple = "ExplorerState.tuple"
    set = "ExplorerState.set"


def get_state(cached_obj: CachedObject):
    if isinstance(cached_obj.obj, dict):
        return ExplorerState.dict
    elif isinstance(cached_obj.obj, list):
        return ExplorerState.list
    elif isinstance(cached_obj.obj, tuple):
        return ExplorerState.tuple
    elif isinstance(cached_obj.obj, set):
        return ExplorerState.set
    else:
        return ExplorerState.public


class Explorer:
    """
    Class representing the explorer object on the left hand side
    This class contains references to the filter and stack objects

    ┌──────────┐ ┌──────────────────┐
    │          │ │                  │
    │          │ │                  │
    │ Explorer │ │     Overview     │
    │ (Filter) │ │                  │
    │ (Stack)  │ │                  │
    └──────────┘ └──────────────────┘
    """

    def __init__(
        self,
        cached_obj: CachedObject,
        term: Terminal,
        filter: Optional[Filter] = None,
        stack: Optional[Stack] = None,
        state: Optional[str] = None,
        public_index: int = 0,
        public_window: int = 0,
        private_index: int = 0,
        private_window: int = 0,
        dict_index: int = 0,
        dict_window: int = 0,
        list_index: int = 0,
        list_window: int = 0,
    ):
        self.cached_obj = cached_obj
        self.term = term
        self.filter = Filter(term=self.term) if not filter else filter
        self.stack = Stack(head_obj=cached_obj) if not stack else stack
        self.public_index = public_index
        self.public_window = public_window
        self.private_index = private_index
        self.private_window = private_window
        self.dict_index = dict_index
        self.dict_window = dict_window
        self.list_index = list_index
        self.list_window = list_window
        self.extra_width = 0

        if state:
            self.state = state
        else:
            self.state = get_state(self.cached_obj)

    def get_layout(self) -> Layout:
        """ Return the layout of the object explorer. This will be a list of lines representing the object attributes/keys/vals we are exploring """
        explorer_layout = Layout(size=self.layout_width)

        if self.state == ExplorerState.dict:
            top_panel = self.dict_panel

        elif self.state in (ExplorerState.list, ExplorerState.tuple, ExplorerState.set):
            top_panel = self.list_panel

        else:
            top_panel = self.dir_panel

        if self.filter.layout.visible:
            combined_layout = Layout()
            combined_layout.split_column(
                top_panel, self.filter.get_layout(self.text_width)
            )
            explorer_layout.update(combined_layout)
        elif self.stack.layout.visible:
            combined_layout = Layout()
            combined_layout.split_column(
                top_panel,
                self.stack.get_layout(
                    width=self.text_width, current_obj=self.cached_obj
                ),
            )
            explorer_layout.update(combined_layout)
        else:
            explorer_layout.update(top_panel)

        return explorer_layout

    @property
    def dir_panel(self) -> Panel:
        lines = []

        if self.state == ExplorerState.public:
            # Reset the public index / window in case applying a filter has now moved the index
            # farther down than it can access on the filtered attributes
            if self.public_index >= len(self.cached_obj.filtered_public_attributes):
                self.public_index = max(
                    0, len(self.cached_obj.filtered_public_attributes) - 1
                )
                self.public_window = max(0, self.public_index - self.num_lines)

            for index, (attr, cached_obj) in enumerate(
                self.cached_obj.filtered_public_attributes.items()
            ):
                line = cached_obj.text.copy()
                if index == self.public_index:
                    line.style += Style(reverse=True)  # type: ignore

                # dim_typeof = cached_obj.typeof.copy()
                # dim_typeof.style = Style(dim=True)
                # line = (
                #     line
                #     + Text(" " * max(2, panel_width - (len(line) + len(cached_obj.typeof))))
                #     + dim_typeof
                # )

                line.truncate(self.text_width)
                lines.append(line)

            title = "[i][cyan]dir[/cyan]()[/i] | [u]public[/u] [dim]private[/dim]"
            subtitle_help = "[dim][u][][/u]:switch pane [/dim]"
            subtitle_index = (
                f"[white]([/white][magenta]{self.public_index + 1 if self.cached_obj.filtered_public_attributes else 0}"
                f"[/magenta][white]/[/white][magenta]{len(self.cached_obj.filtered_public_attributes)}[/magenta][white])"
            )
            if (
                len(console.render_str(subtitle_help + subtitle_index))
                >= self.text_width - 2
            ):
                subtitle = subtitle_index
            else:
                subtitle = subtitle_help + subtitle_index
            if lines == []:
                lines.append(
                    Text("No public attributes", style=Style(color="red", italic=True))
                )

            lines = lines[self.public_window : self.public_window + self.num_lines + 1]

        elif self.state == ExplorerState.private:
            # Reset the private index / window in case applying a filter has now moved the index
            # farther down than it can access on the filtered attributes
            if self.private_index >= len(self.cached_obj.filtered_private_attributes):
                self.private_index = max(
                    0, len(self.cached_obj.filtered_private_attributes) - 1
                )
                self.private_window = max(0, self.private_index - self.num_lines)

            for index, (attr, cached_obj) in enumerate(
                self.cached_obj.filtered_private_attributes.items()
            ):
                line = cached_obj.text.copy()
                if index == self.private_index:
                    line.style += Style(reverse=True)  # type: ignore

                # TODO add a toggle able feature for this
                # dim_typeof = cached_obj.typeof.copy()
                # dim_typeof.style = Style(dim=True)
                # line = (
                #     line
                #     + Text(" " * max(2, panel_width - (len(line) + len(cached_obj.typeof))))
                #     + dim_typeof
                # )

                line.truncate(self.text_width)
                lines.append(line)

            title = "[i][cyan]dir[/cyan]()[/i] | [dim]public[/dim] [u]private[/u]"
            subtitle = (
                "[dim][u][][/u]:switch pane [/dim]"
                f"[white]([/white][magenta]{self.private_index + 1 if self.cached_obj.filtered_private_attributes else 0}"
                f"[/magenta][white]/[/white][magenta]{len(self.cached_obj.filtered_private_attributes)}[/magenta][white])"
            )
            if lines == []:
                lines.append(
                    Text("No private attributes", style=Style(color="red", italic=True))
                )

            lines = lines[self.private_window : self.private_window + self.num_lines]

        if self.num_hidden_attributes:
            num_filtered_line = (
                Text(
                    "+",
                    style=Style(color="white", dim=True, italic=True, underline=True),
                )
                + Text(
                    str(self.num_hidden_attributes),
                    style=Style(color="cyan", dim=True, italic=True),
                )
                + Text(" filtered", style=Style(color="white", dim=True, italic=True))
            )
            num_filtered_line.truncate(self.text_width)
            lines.append(num_filtered_line)

        renderable = Text("\n").join(lines)

        # If terminal is too small don't show the 'dir()' part of the title
        if self.text_width < len(console.render_str(title)) + 3:
            title = title.split("|")[-1].strip()
        if len(console.render_str(title)) > self.text_width:
            if self.state == ExplorerState.public:
                title = "[u]public"
            elif self.state == ExplorerState.private:
                title = "[u]private"

        return Panel(
            renderable,
            title=title,
            title_align="right",
            subtitle=subtitle,
            subtitle_align="right",
            style="white",
            box=box_type,
        )

    @property
    def dict_panel(self) -> Panel:
        """ Return the dictionary explorer layout """

        # Reset the dict index / window in case applying a filter has now moved the index
        # farther down than it can access on the filtered attributes
        if self.dict_index >= len(self.cached_obj.filtered_dict):
            self.dict_index = max(0, len(self.cached_obj.filtered_dict) - 1)
            self.dict_window = max(0, self.dict_index - self.num_lines)

        lines = []

        if self.dict_window == 0:
            lines.append(Text("{"))
            start = 0
            num_lines = self.num_lines - 1
        elif self.dict_window == 1:
            start = 0
            num_lines = self.num_lines
        else:
            start = self.dict_window - 1
            num_lines = self.num_lines

        end = start + num_lines
        index = start

        for attr, (line, cached_obj) in list(self.cached_obj.filtered_dict.items())[
            start:end
        ]:
            new_line = line.copy()
            if index == self.dict_index:
                new_line.style = Style(reverse=True)

            new_line.truncate(self.text_width)
            lines.append(new_line)
            index += 1

        if len(lines) == 1:
            lines[0] = Text("{}")
        else:
            lines.append(Text("}"))
        if self.num_hidden_attributes:
            num_filtered_line = (
                Text(
                    "+",
                    style=Style(color="white", dim=True, italic=True, underline=True),
                )
                + Text(
                    str(self.num_hidden_attributes),
                    style=Style(color="cyan", dim=True, italic=True),
                )
                + Text(" filtered", style=Style(color="white", dim=True, italic=True))
            )
            num_filtered_line.truncate(self.text_width)
            lines.append(num_filtered_line)

        text = Text("\n").join(lines)

        return Panel(
            text,
            title="[i][cyan]dict[/cyan]()",
            title_align="right",
            subtitle=f"([magenta]{self.dict_index + 1}[/magenta]/[magenta]{len(self.cached_obj.filtered_dict)}[/magenta])",
            subtitle_align="right",
            style="white",
            box=box_type,
        )

    @property
    def list_panel(self) -> Panel:
        """ TODO """
        # Reset the list index / window in case applying a filter has now moved the index
        # farther down than it can access on the filtered attributes
        if self.list_index >= len(self.cached_obj.filtered_list):
            self.list_index = max(0, len(self.cached_obj.filtered_list) - 1)
            self.list_window = max(0, self.list_index - self.num_lines)

        lines = []

        bracket_map = {
            ExplorerState.list: ["[", "]", "list"],
            ExplorerState.tuple: ["(", ")", "tuple"],
            ExplorerState.set: ["{", "}", "set"],
        }

        if self.list_window == 0:
            lines.append(Text(bracket_map[self.state][0]))
            start = 0
            num_lines = self.num_lines - 1
        elif self.list_window == 1:
            start = 0
            num_lines = self.num_lines
        else:
            start = self.list_window - 1
            num_lines = self.num_lines

        end = start + num_lines
        index = start

        for line, cached_obj in self.cached_obj.filtered_list[start:end]:
            new_line = line.copy()

            if index == self.list_index:
                new_line.style = Style(reverse=True)

            new_line.truncate(self.text_width)
            lines.append(new_line)
            index += 1

        if len(lines) == 1:
            lines[0] = Text("".join(bracket_map[self.state][:-1]))
        else:
            lines.append(Text(bracket_map[self.state][1]))

        if self.num_hidden_attributes:
            num_filtered_line = (
                Text(
                    "+",
                    style=Style(color="white", dim=True, italic=True, underline=True),
                )
                + Text(
                    str(self.num_hidden_attributes),
                    style=Style(color="cyan", dim=True, italic=True),
                )
                + Text(" filtered", style=Style(color="white", dim=True, italic=True))
            )
            num_filtered_line.truncate(self.text_width)
            lines.append(num_filtered_line)

        text = Text("\n").join(lines)

        return Panel(
            text,
            title=f"[i][cyan]{bracket_map[self.state][2]}[/cyan]()",
            title_align="right",
            subtitle=f"([magenta]{self.list_index + 1}[/magenta]/[magenta]{len(self.cached_obj.filtered_list)}[/magenta])",
            subtitle_align="right",
            style="white",
            box=box_type,
        )

    def explore_selected_object(self) -> Optional[CachedObject]:
        """ TODO """

        # Save current stack as a frame
        current_frame = StackFrame(
            cached_obj=self.cached_obj,
            filter=self.filter,
            state=self.state,
            public_index=self.public_index,
            public_window=self.public_window,
            private_index=self.private_index,
            private_window=self.private_window,
            dict_index=self.dict_index,
            dict_window=self.dict_window,
            list_index=self.list_index,
            list_window=self.list_window,
        )
        self.stack.push(current_frame)

        self.cached_obj = self.selected_object
        self.cached_obj.cache()
        self.state = get_state(self.cached_obj)
        self.filter = Filter(term=self.term)
        self.public_index = 0
        self.public_window = 0
        self.private_index = 0
        self.private_window = 0
        self.dict_index = 0
        self.dict_window = 0
        self.list_index = 0
        self.list_window = 0

        return None

    def explore_parent_obj(self):
        """ Go back to exploring the parent obj of the current obj """
        stack_frame = self.stack.pop()
        if stack_frame:
            self.cached_obj = stack_frame.cached_obj
            self.filter = stack_frame.filter
            self.state = stack_frame.state
            self.public_index = stack_frame.public_index
            self.public_window = stack_frame.public_window
            self.private_index = stack_frame.private_index
            self.private_window = stack_frame.private_window
            self.dict_index = stack_frame.dict_index
            self.dict_window = stack_frame.dict_window
            self.list_index = stack_frame.list_index
            self.list_window = stack_frame.list_window

        return self.cached_obj

    def explore_selected_stack_object(self):
        stack_frame = self.stack.select()
        if stack_frame:
            self.cached_obj = stack_frame.cached_obj
            self.filter = stack_frame.filter
            self.state = stack_frame.state
            self.public_index = stack_frame.public_index
            self.public_window = stack_frame.public_window
            self.private_index = stack_frame.private_index
            self.private_window = stack_frame.private_window
            self.dict_index = stack_frame.dict_index
            self.dict_window = stack_frame.dict_window
            self.list_index = stack_frame.list_index
            self.list_window = stack_frame.list_window

        return self.cached_obj

    def move_up(self):
        """ Move the current selection up one """
        if self.state == ExplorerState.public:
            if self.public_index > 0:
                self.public_index -= 1
                if self.public_index < self.public_window:
                    self.public_window -= 1

        elif self.state == ExplorerState.private:
            if self.private_index > 0:
                self.private_index -= 1
                if self.private_index < self.private_window:
                    self.private_window -= 1

        elif self.state == ExplorerState.dict:
            if self.dict_index > 0:
                self.dict_index -= 1
                if self.dict_index < self.dict_window - 1:
                    self.dict_window -= 1
            elif self.dict_window == 1:
                self.dict_window -= 1

        elif self.state in (ExplorerState.list, ExplorerState.tuple, ExplorerState.set):
            if self.list_index > 0:
                self.list_index -= 1
                if self.list_index < self.list_window - 1:
                    self.list_window -= 1
            elif self.list_window == 1:
                self.list_window -= 1

    def move_down(self):
        """ Move the current selection down one """
        if self.state == ExplorerState.public:
            if self.public_index < self.num_filtered_attributes - 1:
                self.public_index += 1
                if self.public_index >= self.public_window + self.num_lines:
                    self.public_window += 1
            elif (
                self.public_window == self.num_filtered_attributes - self.num_lines + 1
            ):
                self.public_window += 1

        elif self.state == ExplorerState.private:
            if self.private_index < self.num_filtered_attributes - 1:
                self.private_index += 1
                if self.private_index >= self.private_window + self.num_lines:
                    self.private_window += 1
            elif (
                self.private_window == self.num_filtered_attributes - self.num_lines + 1
            ):
                self.private_window += 1

        elif self.state == ExplorerState.dict:
            if self.dict_index < self.num_filtered_attributes - 1:
                self.dict_index += 1
                if self.dict_index >= self.dict_window + self.num_lines - 1:
                    self.dict_window += 1
            elif self.dict_window == self.num_filtered_attributes - self.num_lines + 1:
                self.dict_window += 1
            elif (
                self.dict_window == self.num_filtered_attributes - self.num_lines + 2
                and self.num_hidden_attributes > 0
            ):
                self.dict_window += 1

        elif self.state in (ExplorerState.list, ExplorerState.tuple, ExplorerState.set):
            if self.list_index < self.num_filtered_attributes - 1:
                self.list_index += 1
                if self.list_index >= self.list_window + self.num_lines - 1:
                    self.list_window += 1
            elif self.list_window == self.num_filtered_attributes - self.num_lines + 1:
                self.list_window += 1
            elif (
                self.list_window == self.num_filtered_attributes - self.num_lines + 2
                and self.num_hidden_attributes > 0
            ):
                self.list_window += 1

    def move_top(self):
        if self.state == ExplorerState.public:
            self.public_index = 0
            self.public_window = 0

        elif self.state == ExplorerState.private:
            self.private_index = 0
            self.private_window = 0

        elif self.state == ExplorerState.dict:
            self.dict_index = self.dict_window = 0

        elif self.state in (ExplorerState.list, ExplorerState.tuple, ExplorerState.set):
            self.list_index = self.list_window = 0

    def move_bottom(self):
        """Move all the way to the bottom. If there are hidden attributes, make sure to show that line by
        increasing the window index by 1"""
        if self.state == ExplorerState.public:
            self.public_index = self.num_filtered_attributes - 1
            self.public_window = max(
                0,
                self.public_index
                - self.num_lines
                + (1 if self.num_hidden_attributes == 0 else 2),
            )
        elif self.state == ExplorerState.private:
            self.private_index = self.num_filtered_attributes - 1
            self.private_window = max(
                0,
                self.private_index
                - self.num_lines
                + (1 if self.num_hidden_attributes == 0 else 2),
            )
        elif self.state == ExplorerState.dict:
            self.dict_index = self.num_filtered_attributes - 1
            self.dict_window = max(
                0,
                self.dict_index
                - self.num_lines
                + (3 if self.num_hidden_attributes == 0 else 4),
            )
        elif self.state in (ExplorerState.list, ExplorerState.tuple, ExplorerState.set):
            self.list_index = self.num_filtered_attributes - 1
            self.list_window = max(
                0,
                self.list_index
                - self.num_lines
                + (3 if self.num_hidden_attributes == 0 else 4),
            )

    def copy(self):
        return Explorer(
            term=self.term,
            cached_obj=self.cached_obj,
            filter=self.filter,
            stack=self.stack,
            state=self.state,
            public_index=self.public_index,
            public_window=self.public_window,
            private_index=self.private_index,
            private_window=self.private_window,
            dict_index=self.dict_index,
            dict_window=self.dict_window,
            list_index=self.list_index,
            list_window=self.list_window,
        )

    def reset_index(self):
        if self.public_index >= self.num_filtered_attributes:
            self.public_index = self.num_filtered_attributes - 1
            self.public_window = max(0, self.public_index - self.num_lines + 2)
        elif self.public_index < 0:
            self.public_index = 0
        if self.private_index >= self.num_filtered_attributes:
            self.private_index = self.num_filtered_attributes - 1
            self.private_window = max(0, self.private_index - self.num_lines + 2)
        elif self.private_index < 0:
            self.private_index = 0
        if self.dict_index >= self.num_filtered_attributes:
            self.dict_index = self.num_filtered_attributes - 1
            self.dict_window = max(0, self.dict_index - self.num_lines + 4)
        elif self.dict_index < 0:
            self.dict_index = 0
        if self.list_index >= self.num_filtered_attributes:
            self.list_index = self.num_filtered_attributes - 1
            self.list_window = max(0, self.list_index - self.num_lines + 4)
        elif self.list_index < 0:
            self.list_index = 0

    @property
    def num_attributes(self) -> int:
        """ Return the number of attributes of the current cached object """
        if self.state == ExplorerState.public:
            return self.cached_obj.num_public_attributes
        elif self.state == ExplorerState.private:
            return self.cached_obj.num_private_attributes
        else:
            return self.cached_obj.length or 0

    @property
    def num_filtered_attributes(self) -> int:
        """ Return the number of filtered attributes """
        if self.state == ExplorerState.public:
            return self.cached_obj.num_filtered_public_attributes
        elif self.state == ExplorerState.private:
            return self.cached_obj.num_filtered_private_attributes
        elif self.state == ExplorerState.dict:
            return self.cached_obj.num_filtered_dict_keys
        else:
            return self.cached_obj.num_filtered_list_items

    @property
    def num_hidden_attributes(self) -> int:
        return self.num_attributes - self.num_filtered_attributes

    @property
    def num_lines(self):
        return self.term.height - 5

    @property
    def live_update(self) -> bool:
        """True/False value wheter to live update the filters of the cached object
        If the number of visible attributes is over a threshold we do not live update
        the search filter
        """
        return self.num_attributes < 130

    @property
    def selected_object(self) -> CachedObject:
        """ Return the currently selected cached object """
        try:
            if self.state == ExplorerState.public:
                attr = list(self.cached_obj.filtered_public_attributes.keys())[
                    self.public_index
                ]
                return self.cached_obj.filtered_public_attributes[attr]

            elif self.state == ExplorerState.private:
                attr = list(self.cached_obj.filtered_private_attributes.keys())[
                    self.private_index
                ]
                return self.cached_obj.filtered_private_attributes[attr]

            elif self.state == ExplorerState.dict:
                # Get the currently selected key
                key = list(self.cached_obj.filtered_dict)[self.dict_index]
                return self.cached_obj.filtered_dict[key].cached_object

            elif self.state in (
                ExplorerState.list,
                ExplorerState.tuple,
                ExplorerState.set,
            ):
                return self.cached_obj.filtered_list[self.list_index][1]
            else:
                raise ValueError("Unexpected explorer state")

        except (KeyError, IndexError):
            return CachedObject(None)

    @property
    def layout_width(self):
        layout_width = (self.term.width - 2) // 4 + self.extra_width
        if layout_width > self.term.width - 20:
            layout_width = self.term.width - 20
            self.extra_width = 0
        return layout_width

    @property
    def text_width(self):
        """ Return the width of text allowed within the panel """
        return self.layout_width - 4

    def increase_width(self):
        if self.layout_width < self.term.width - 20:
            self.extra_width += 1

    def decrease_width(self):
        if self.layout_width + self.extra_width > 11:
            self.extra_width -= 1
